#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright(c) 2011 Krister Hedfors
#
# This file is part of pyTickets.
#
#    pyTickets is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    pyTickets is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with pyTickets.  If not, see <http://www.gnu.org/licenses/>.
#



class Interface(object): pass
class NotImplementedException(Exception): pass


class ISerializable(Interface):
    def serialize(self):
        raise NotImplementedException('implement serialize()')
        
    @classmethod
    def deserialize(self, b):
        raise NotImplementedException('implement @classmethod deserialize()')



class Serializer(object):
    '''
        Separate Serializer class, not indended to be subclassed by classes
        wishing to be serialized. Serializer instances are intended to be used
        by MultiSerializable instances.
    '''

    def serialize(self, val):
        raise NotImplementedException('implement serialize()')
        
    def deserialize(self, val):
        raise NotImplementedException('implement @classmethod deserialize()')



class MultiSerializable(ISerializable):

    _serializers = []

    @classmethod
    def add_serializer(cls, serializer):
        cls._serializers.append(serializer)

    def serialize(self):
        val = self
        for serializer in self._serializers:
            val = serializer.serialize(val)
        return val

    @classmethod
    def deserialize(cls, val):
        serializers = cls._serializers
        for i in xrange(len(serializers)-1, -1, -1):
            val = serializers[i].deserialize(val)
        return cls(val)
    
    

import doctest
import unittest
class Test_MultiSerializable(unittest.TestCase):

    def test_test(self):
        assert 1==1

    def test1(self):

        class MS(list, MultiSerializable):
            pass

        class AddOne(Serializer):
            def serialize(self, val):
                lst = val[:]
                for i in xrange(len(lst)):
                    lst[i] += 1
                return lst
            def deserialize(self, val):
                lst = val[:]
                for i in xrange(len(lst)):
                    lst[i] -= 1
                return lst

        class AddTwo(Serializer):
            def serialize(self, val):
                lst = val[:]
                for i in xrange(len(lst)):
                    lst[i] += 2
                return lst
            def deserialize(self, val):
                lst = val[:]
                for i in xrange(len(lst)):
                    lst[i] -= 2
                return lst

        MS.add_serializer(AddOne())
        MS.add_serializer(AddTwo())

        ms = MS([0,1,2])
        assert len(ms) == 3

        ms = ms.serialize()
        assert ms == [3,4,5]
        ms = MS.deserialize(ms)
        assert ms == [0,1,2]
        


def _test():
    doctest.testmod()
    unittest.main()

if __name__ == "__main__":
    _test()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright(c) 2011 Krister Hedfors
#
# This file is part of pyTickets.
#
#    pyTickets is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    pyTickets is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with pyTickets.  If not, see <http://www.gnu.org/licenses/>.
#


from tickets import SecureTicket
from tickets import SecureTicketPublicFlags
from tickets import SecureTicketFlags
from tickets import SecureTicketService
from tickets import SecureTicketException
from tickets import InvalidTicketException

from tickets.tests.utils import flip_bit

import unittest
import random
import uuid
import zlib
import datetime

try:
    from sets import Set as set
except ImportError:
    pass


def debug_print_ticket(ticket):
    print ticket
    for name in ('_bytes', 'prefix', 'hash', 'salt', 'raw_public_flags', 'raw_flags',
                'raw_valid_until', 'raw_data'):
        value =  getattr(ticket, name)
        print len(value), name, repr(value)



class Test_new_SecureTicketService(unittest.TestCase):
    def _test_test(self):
            assert 1==2

    def setUp(self):
        pass

    def test_various_combinations_and_bitflips(self):
        key_variations = [
            SecureTicketService.create_random_key() for _ in xrange(2)
        ]
        kw_variations = [
            dict(encrypt=0, serialize=0, compress=0),
            dict(encrypt=0, serialize=0, compress=1),
            dict(encrypt=0, serialize=1, compress=0),
            dict(encrypt=0, serialize=1, compress=1),
            dict(encrypt=1, serialize=0, compress=0),
            dict(encrypt=1, serialize=0, compress=1),
            dict(encrypt=1, serialize=1, compress=0),
            dict(encrypt=1, serialize=1, compress=1),
        ]
        entropy_variations = [
            'E'*i for i in (0, 1, 15, 16, 17)
        ]
        data_variations = [
            'D'*i for i in (0, 1, 15, 16, 17)
        ]

        for key in key_variations:
            for kw in kw_variations:
                for entropy in entropy_variations:
                    for data in data_variations:
                        #print kw

                        #print 'data, entropy:', repr(entropy), repr(data)

                        sts = SecureTicketService(key, **kw)

                        t1 = sts.create_ticket(data, entropy)
                        t2 = SecureTicket.fromstring(t1.tostring())
                        t3 = SecureTicket.frombase64(t2.tobase64())

                        assert sts.validate_ticket(t1, entropy)
                        if not t1.is_encrypted:
                            assert t1.data == data
                        assert sts(t1).data == data

                        assert sts.validate_ticket(t2, entropy)
                        if not t2.is_encrypted:
                            assert t2.data == data
                        assert sts(t2).data == data

                        assert sts.validate_ticket(t3, entropy)
                        if not t3.is_encrypted:
                            assert t3.data == data
                        assert sts(t3).data == data

                        s = t1.tostring()
                        flip_pos = random.randint(0, len(s)*8-1)
                        t = flip_bit(s, flip_pos)
                        assert s != t
                        try:
                            t4 = SecureTicket.fromstring(t)
                            assert not sts.validate_ticket(t4, entropy)
                        except InvalidTicketException:
                            pass

                        u = flip_bit(t, flip_pos)
                        t5 = SecureTicket.fromstring(u)
                        assert sts.validate_ticket(t5, entropy)

                        #if not sts(ticket).data == data:
                        #    print 'FAIL data==data', ticket


    def test_serialization(self):
        key_variations = [
            SecureTicketService.create_random_key() for _ in xrange(2)
        ]
        kw_variations = [
            dict(encrypt=0, serialize=1, compress=0),
            dict(encrypt=0, serialize=1, compress=1),
            dict(encrypt=1, serialize=1, compress=0),
            dict(encrypt=1, serialize=1, compress=1),
        ]
        entropy_variations = [
            '', 'EEEEEEEEEEE'
        ]
        data_variations = [
            None,
            [],
            (),
            dict(),
            1,
            2L,
            3.14,
            set(),
            [
                123,
                [
                    dict(asd=123, qwe=456),
                    ['foo', ('bar','qwe')],
                    set(['robot', u'brains', None])
                ]
            ]
        ]
        for key in key_variations:
            for entropy in entropy_variations:
                for data in data_variations:
                    for kw in kw_variations:
                        sts = SecureTicketService(key, **kw)
                        t1 = sts.create_ticket(data, entropy)
                        assert sts.validate_ticket(t1, entropy)
                        if not t1.is_encrypted:
                            assert t1.data == data
                        assert sts(t1).data == data
                        #if t1.ticket_len > 100:
                        #    print '--------------------------------------'
                        #    print t1.ticket_len, kw
                        #    print data


    def test_large_tickets(self):
        # these take time
        kw_variations = [
            #dict(encrypt=0, serialize=0, compress=0),
            #dict(encrypt=0, serialize=0, compress=1),
            #dict(encrypt=0, serialize=1, compress=0),
            #dict(encrypt=0, serialize=1, compress=1),
            #dict(encrypt=1, serialize=0, compress=0),
            #dict(encrypt=1, serialize=0, compress=1),
            #dict(encrypt=1, serialize=1, compress=0),
            dict(encrypt=1, serialize=1, compress=1),
        ]
        for kw in kw_variations:
            from tickets.log import debug, info
            key = SecureTicketService.create_random_key()
            sts = SecureTicketService(key, **kw)
            data = 'D' * (10**6)
            ticket = sts.create_ticket(data, 'asdasd')
            if not kw['encrypt']:
                assert ticket.data == data
            assert sts(ticket).data == data
            assert sts.validate_ticket(ticket, 'asdasd')
            assert not sts.validate_ticket(ticket, 'asdasd.')
    

class Test_SecureTicket(unittest.TestCase):
    def _test_test(self):
            assert 1==1

    def setUp(self):
        pass

    def test_properties_and_serialization(self):
        salt_variations = [
            uuid.uuid4().bytes[:SecureTicket.SALT_LEN] for _ in xrange(4)
        ]
        hash_variations = [
            (uuid.uuid4().bytes*2)[:SecureTicket.HASH_LEN] for _ in xrange(4)
        ]
        public_flags_variations = [
            SecureTicketPublicFlags(encrypt=0),
            #SecureTicketPublicFlags(encrypt=1),
        ]
        flags_variations = [
            SecureTicketFlags(serialize=0, compress=0)
        ]
        valid_until_variations = [
            random.randint(0, 0xffffffff) for _ in xrange(4)
        ]
        data_variations = [
            'X'*i for i in xrange(20)
        ]
        for salt in salt_variations:
            for hash_ in hash_variations:
                for public_flags in public_flags_variations:
                    for flags in flags_variations:
                        for valid_until in valid_until_variations:
                            for data in data_variations:
                                ticket = SecureTicket()._load_from_parts(
                                    hash_,
                                    salt,
                                    public_flags, 
                                    flags,
                                    valid_until,
                                    data
                                )
                                if ticket.raw_valid_until == '':
                                    print 'BROKEN', repr(valid_until), valid_until
                                    debug_print_ticket(ticket)
                                    raise Exception
                                assert len(ticket.raw_valid_until) == 4
                                assert ticket.salt         == salt
                                assert ticket.hash         == hash_
                                #
                                # SecureTicketFlags unpacks into new instances.
                                #
                                assert ticket.flags._serialize() == flags._serialize()
                                assert ticket.valid_until        == valid_until 
                                assert ticket.data               == data



class Test_SecureTicketService(unittest.TestCase):
    def _test_test(self):
        assert 1==1

    def setUp(self):
        self.sts = SecureTicketService('some key')
        self.sts_other = SecureTicketService('other key')

    def test_validate(self):
        '''
            Validate tickets created from all combinations of random keys of length 1 to 16 bytes,
            entropy of length 0 to 16 and data of length 1 to 16.

            Assert validation failure for each successful validation, when using a key which
            differs by one bit from the correct key.

        '''
        key_variations = [
            uuid.uuid4().bytes[:i] for i in xrange(1,16)
        ]
        entropy_variations = [
            uuid.uuid4().bytes[:i] for i in xrange(8)
        ]
        data_variations = [
            'A'*i for i in xrange(16,64)
        ]
        key_flip_bit = 0
        for key in key_variations:
            sts = SecureTicketService(key)
            stsfail = SecureTicketService(
                flip_bit(key, key_flip_bit % (len(key)*8))
            )
            key_flip_bit += 7

            for entropy in entropy_variations:
                for data in data_variations:
                    assert not data is None
                    ticket = sts.create_ticket(data, entropy)
                    assert sts.validate_ticket(ticket, entropy)
                    assert not stsfail.validate_ticket(ticket, entropy)

    def test_data(self):
        key_variations = [
            uuid.uuid4().bytes[:i] for i in xrange(1,16)
        ]
        entropy_variations = [
            uuid.uuid4().bytes[:i] for i in xrange(8)
        ]
        data_variations = [
            'A'*i for i in xrange(16,33)
        ]
        key_flip_bit = 0
        for key in key_variations:
            sts = SecureTicketService(key)
            stsfail = SecureTicketService(
                flip_bit(key, key_flip_bit % (len(key)*8))
            )
            key_flip_bit += 7

            for entropy in entropy_variations:
                for data in data_variations:
                    assert not data is None
                    ticket = sts.create_ticket(data, entropy)
                    assert sts.validate_ticket(ticket, entropy)
                    assert sts(ticket).data == data

                



def choffset(self, string, offset, val):
    import array
    arr = array.array('c', string)
    newval = chr( (ord(arr[offset])+val) % 256 )
    arr[offset] = newval
    return arr.tostring()



__all__ = []
for (name,val) in locals().copy().iteritems():
    try:
        if issubclass(val, unittest.TestCase):
            __all__.append(name)
    except:
        pass


def _test():
    unittest.main()

if __name__ == "__main__":
    _test()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright(c) 2011 Krister Hedfors
#
# This file is part of pyTickets.
#
#    pyTickets is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    pyTickets is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with pyTickets.  If not, see <http://www.gnu.org/licenses/>.
#


__all__ = [
    'ByteMapperFactory'
]




class BaseByteMapper(object):

    @classmethod
    def _join_strings(cls, names_values):
        def string_start_offset(name):
            return getattr(cls, name.upper()+'_OFFSET')
        sorted_strings = [
            names_values[name] for name in
                sorted(names_values, key=string_start_offset)
        ]
        return ''.join(sorted_strings)

    @classmethod
    def _serialize_join_strings(cls, names_values, **kw):
        _names_values = {}
        for (name,val) in names_values.iteritems():
            serialize = getattr(cls, '_serialize_'+name)
            val = serialize(val, **kw)
            assert type(val) is str
            _names_values[name] = val
        names_values = _names_values
        return cls._join_strings(names_values)

    @classmethod
    def _split_strings(cls, plaintext, *names):
        '''
            beware: no alerts if plaintext is too small
        '''
        def string_start_offset(name):
            return getattr(cls, name.upper()+'_OFFSET')
        sorted_strings = []
        for name in sorted(names, key=string_start_offset):
            len_ = getattr(cls, name.upper()+'_LEN')
            s = plaintext[:len_]
            sorted_strings.append(s)
            plaintext = plaintext[len_:]
        return tuple(sorted_strings)

    @classmethod
    def _split_deserialize_strings(cls, plaintext, names):
        '''
            beware: no alerts if plaintext is too small
        '''
        def string_start_offset(name):
            return getattr(cls, name.upper()+'_OFFSET')
        sorted_values = []
        for name in sorted(names, key=string_start_offset):
            len_ = getattr(cls, name.upper()+'_LEN')
            deserialize = getattr(cls, '_deserialize_'+name)
            s = deserialize(plaintext[:len_])
            sorted_values.append(s)
            plaintext = plaintext[len_:]
        return tuple(sorted_values)




class ByteMapperFactory(object):
    '''
        ByteMapperFactory creates new classes able to parse custom byte ranges
        from attribute _bytes. Parsing is done by slicing. Attribute _bytes can
        be of any sliceable type but must be explicitly set in instances!

        Make sure to declare _bytes, preferrably in __init__() of inheriting classes:

        >>> _MyBM = ByteMapperFactory('_MyBM', [ (3,'asd'), (3,'qwe') ])
        >>> class MyBM(_MyBM):
        ...     def __init__(self):
        ...         super(MyBM, self).__init__()
        ...         self._bytes = [1,2,3,4,5,6]
        ...
        >>> a = MyBM()
        >>> assert a.asd == [1,2,3]
        >>> assert a.qwe == [4,5,6]
        >>> assert a.QWE_OFFSET == 3

    '''
    @classmethod
    def classmethod_parse(cls, start, end):
        @classmethod
        def parse(cls, b):
            return b[start : end]
        return parse

    @classmethod
    def classmethod_serialize(cls):
        @classmethod
        def serialize(self, val, **kw):
            return val
        return serialize

    @classmethod
    def classmethod_deserialize(cls):
        @classmethod
        def deserialize(self, val, **kw):
            return val
        return deserialize

    @classmethod
    def property_get(cls, name):
        @property
        def getter(self):
            parse = getattr(self.__class__, '_parse_'+name)
            deserialize = getattr(self.__class__, '_deserialize_'+name)
            val = parse(self._bytes)
            return deserialize(val)
        return getter


    def __new__(cls, classname, sizes_names):
        d = {}
        d['_ByteMapperFactory_args'] = (classname, sizes_names)
        d['_offsets'] = {}
        offset = 0
        for tup in sizes_names:
            try:
                size, name, docs = tup
            except ValueError:
                size, name = tup
                docs = ''

            start, end = offset, offset+size
            #
            # class attributes
            #
            d['_offsets'][name]       = (start, end)
            d[name.upper()+'_OFFSET'] = start
            d[name.upper()+'_LEN']    = size
            d['OFFSET_'+name.upper()] = start
            #
            # classmethods
            #
            d['_parse_'+name]               = cls.classmethod_parse(start, end)
            d['_serialize_'+name]           = cls.classmethod_serialize()
            d['_deserialize_'+name]         = cls.classmethod_deserialize()
            #
            # properties
            #
            d[name]                   = cls.property_get(name)
            offset += size
        d['_bytes_len'] = offset
        return type(classname, (BaseByteMapper,), d)




import unittest
class Test_ByteMapperFactory(unittest.TestCase):
    def test_test(self):
        assert 1==1

    def test1(self):
        X = ByteMapperFactory('X', [
            (2, 'asd'),
            (2, 'qwe'),
            (4, 'foo'),
            (4, 'bar'),
        ])

        class Y(X):
            def __init__(self):
                super(Y, self).__init__()
                self._bytes = range(self._bytes_len)

        y = Y()
        assert y.asd == [0,1]
        assert y.qwe == [2,3]
        assert y.foo == [4,5,6,7]
        assert y.bar == [8,9,10,11]

        class Z(Y):
            @classmethod
            def _deserialize_asd(cls, val):
                return [str(i) for i in val]

        z = Z()
        assert z.asd == ['0', '1']
        assert z.qwe == [2, 3]

    def test_join_strings(self):
        X = ByteMapperFactory('X', [
            (2, 'asd'),
            (2, 'qwe'),
            (4, 'foo'),
            (4, 'bar'),
        ])
        res = X._join_strings(dict(qwe='QWE', bar='BAR', asd='ASD'))
        assert res == 'ASDQWEBAR'

        class Y(X):
            @classmethod
            def _serialize_bar(cls, val):
                return val+'rrr'
        res = Y._serialize_join_strings(dict(qwe='QWE', bar='BAR', asd='ASD'))
        assert res == 'ASDQWEBARrrr'


        


def _test():
    import doctest
    import unittest
    doctest.testmod()
    unittest.main()

if __name__ == "__main__":
    _test()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright(c) 2011 Krister Hedfors
#
#
# This file is part of pyTickets.
#
#    pyTickets is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    pyTickets is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with pyTickets.  If not, see <http://www.gnu.org/licenses/>.
#


__all__ = [
    'FlagsFactoryException',
    'FlagsFactory',
]


import struct
import operator

try:
    from sets import Set as set
except ImportError:
    pass
    




class FlagsFactoryException(Exception):
    pass

class FlagsException(Exception):
    pass


class FlagsBase(object):

    _ignore_invalid_names_on_init = False

    def __init__(self, flags=None, **kw):
        if not flags is None:
            assert type(flags) is long
            self._flags = flags
        for (name,val) in kw.iteritems():
            if not name in self._flagbit:
                if not self._ignore_invalid_names_on_init:
                    msg = 'unknown flag name: "%s"' %name
                    raise FlagsFactoryException(msg)
            else:
                if val:
                    self._flags |= self._flagbit[name]
                else:
                    self._flags &= ~self._flagbit[name]

    def __getitem__(self, name):
        name = name.lower()
        try:
            idx = operator.indexOf(self._flag_names, name)
        except ValueError:
            msg = "Invalid flag name: '{0}'.".format(name)
            msg += " Valid names are: {0}.".format(', '.join(self._flag_names))
            raise FlagsException(msg)
        bit = 1 << idx
        return bool(self._flags & bit)

    def _set(self, name, value):
        if not name in self._flagbit:
            msg = "Unknown flag name: '{0}'".format(name)
            raise FlagsException(msg)
        if value:
            self._flags |= self._flagbit[name]
        else:
            self._flags &= ~self._flagbit[name]

    def set(self, *args, **kw):
        if len(args):
            assert len(args) == 2
            name, value = args
            self._set(name, value)
        for (name, value) in kw.iteritems():
            self._set(name, value)

    @classmethod
    def _deserialize(cls, flags_serialized):
        flags = long(struct.unpack(cls._type, flags_serialized)[0])
        return cls(flags=flags)

    def _serialize(self):
        return struct.pack(self._type, self._flags)

    def subset(self, *names):
        d = {}
        for name in names:
            d[name] = self[name]
        return d

    def _get_d(self):
        d = {}
        for (i,name) in enumerate(self._flag_names):
            bit = 1 << i
            val = bool(self._flags & bit)
            d[name] = val
        return d
    d = property(_get_d)

    def _get_names(self):
        return tuple(self._flag_names)
    names = property(_get_names)

    def sync(self, otherflags):
        '''
            update self with values from otherflags for all matching flag names
        '''
        names = set(self.names) & set(otherflags.names)
        self.set( **otherflags.subset( *names))

    def copy(self):
        return self.__class__(**self.d)

    def copy_update(self, **kw):
        f = self.copy()
        f.set(**kw)
        return f

    def copy_sync(self, otherflags):
        f = self.copy()
        f.sync(otherflags)
        return f


        



class FlagsFactory(object):

    @classmethod
    def _get_type(cls, size):
        _type = None
        if size == 8:
            _type = 'B'
        elif size == 16:
            _type = 'H'
        elif size == 32:
            _type = 'I'
        elif size == 64:
            _type = 'Q'
        else:
            msg = 'size must be one of 8, 16, 32, 64'
            raise FlagsFactoryException(msg)
        return _type
        
    @classmethod
    def _property_get(cls, name):
        def _get(self):
            flagbit = getattr(self, 'F_'+name.upper())
            return bool(self._flags & flagbit)
        return property(_get)

    @classmethod
    def _method_set_name(cls, name):
        def _set(self, value):
            self.set(name, value)
        return _set

    def __new__(cls, classname, size, names_defaults):
        d = {} 
        _flag_names = []
        _flag_names_defaults = []
        _type = ''
        _flagbit = {}

        d['_FlagsFactory_args'] = (classname, size, names_defaults)

        for (name,val) in names_defaults:
            name = name.lower()
            val = bool(val)
            _flag_names.append(name)
            _flag_names_defaults.append((name, val))
        if len(set(_flag_names)) < len(_flag_names):
            msg = 'flag names not unique'
            raise FlagFactoryException(msg)
        d['_flag_names'] = _flag_names
        d['_flag_names_defaults'] = _flag_names_defaults

        if size < len(_flag_names):
            msg = 'number of flag names is larger than size (%d)' % size
            raise FlagsFactoryException(msg)
        d['_type'] = cls._get_type(size)
        d['_size'] = size

        for (i,name) in enumerate(_flag_names):
            val = 1 << i
            _flagbit[name]       = val
            d['F_'+name.upper()] = val
            d['set_'+name]       = cls._method_set_name(name)
            d[name]              = cls._property_get(name)
        d['_flagbit'] = _flagbit

        _flags = 0L
        for (name,default) in _flag_names_defaults:
            if default:
                _flags |= _flagbit[name]
        d['_flags'] = _flags

        return type(classname, (FlagsBase,), d)



import unittest
class Test_FlagsFactory(unittest.TestCase):
    def test_test(self):
        assert 1==1

    def test1(self):
        X = FlagsFactory('X', 8, [
            ('asd', True),
            ('qwe', False),
            ('foo', False),
            ('bar', False),
        ])
        x = X()
        assert x.asd
        assert not x.qwe
        assert not x.foo
        assert not x.bar
        x2 = X(qwe=True)
        assert x2.asd
        assert x2.qwe
        assert not x2.foo
        assert not x2.bar

        x.set('asd', False)
        assert not x.asd
        x.set('asd', True)
        assert x.asd

        x.set_asd(False)
        assert not x.asd
        x.set_asd(True)
        assert x.asd

        x.set(**dict(
            asd=0,
            qwe=0,
            foo=0,
            bar=0,
        ))
        assert x.d == dict(
            asd=False,
            qwe=False,
            foo=False,
            bar=False,
        )

        x.set(**dict(
            asd=1,
            qwe=1,
            foo=1,
            bar=1,
        ))
        assert type(x) is type(X())
        assert x['asd'] == True
        assert x['qwe'] == True
        assert type(x.d) is type({})
        assert x.d['asd'] == True
        assert x.d['qwe'] == True

        assert x.subset('foo', 'bar') == dict(foo=True,bar=True)
        assert x.names == tuple(['asd','qwe','foo','bar'])

        W = FlagsFactory('W', 8, [
            ('asd',   False),
            ('blurg', False),
        ])
        w = W()
        w.sync(x)
        assert w.d == dict(asd=True, blurg=False)


def _test():
    import doctest
    import unittest
    doctest.testmod()
    unittest.main()

if __name__ == "__main__":
    _test()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright(c) 2011 Krister Hedfors:
#
# This file is part of pyTickets.
#
#    pyTickets is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    pyTickets is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with pyTickets.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = [
    #'SecureTicketException',
    #'SecureTicketValidationException',
    #'InvalidTicketException',
    #'SecureTicketPublicFlags',
    #'SecureTicketFlags',
    #'SecureTicket',
    #'SecureTicketService',
]

import struct
import hmac
import hashlib
import uuid
import zlib
import base64
import time
import datetime
import pdb
import pickle
import cPickle

from tickets.bytemapper import ByteMapperFactory
import tickets.crypto
import tickets.flags

#
# Warn if SecureTicket's ticket.valid_until is accessed while being zero, as this usually
# indicates that valid_until is encrypted and needs to be obtained through
# SecureTicketService.get_valid_until(ticket) instead.
#
WARN_EMPTY_VALID_UNTIL = 0



class SecureTicketException(Exception): pass

class SecureTicketValidationException(SecureTicketException): pass
class InvalidTicketException(SecureTicketValidationException): pass




SecureTicketPublicFlags = tickets.flags.FlagsFactory(
    'SecureTicketPublicFlags',
    8,
    [
        ('encrypt', False),
    ]
)

SecureTicketFlags = tickets.flags.FlagsFactory(
    'SecureTicketFlags',
    8,
    [
        #
        # These flags in a ticket informs getter methods wether or not
        # data is serialized and/or compressed.
        #
        ('serialize',         False),
        ('compress',          False),
    ]
)

_SecureTicket = ByteMapperFactory(
    '_SecureTicket',
    [
        (4,  'prefix'),
        (20, 'hash'),
        (16, 'salt'),
        (4,  'ticket_len'),
        (1,  'public_flags'),
        (1,  'flags'),
        (4,  'valid_until'),
        (4,  'data_len'),
    ]
)





class SecureTicket(_SecureTicket):
    '''
    Instances of *SecureTicket* should generally be obtained through
    ``SecureTicketService.create_ticket()``, alternatively through calls to
    the classmethods ``SecureTicket.fromstring()`` or
    ``SecureTicket.frombase64()``.
    '''

    VISIBLE_MESSAGE_OFFSET = _SecureTicket.SALT_OFFSET
    CIPHERTEXT_OFFSET      = _SecureTicket.FLAGS_OFFSET
    DATA_OFFSET            = _SecureTicket.DATA_LEN_OFFSET +\
                                _SecureTicket.DATA_LEN_LEN

    MAX_LEN             = 2**32-1 # too large actualy, ticket uses some bytes
    VISIBLE_MESSAGE_LEN = MAX_LEN
    DATA_LEN            = MAX_LEN

    HASH_TYPE = 'sha256'

    SECURE_TICKET_PREFIX = struct.pack('I', 0x5ec71c37) # "secticet"

    _pickler = cPickle
    #_pickler = pickle

    def __init__(self, **kw):
        self._kw = kw

    @classmethod
    def _create_plaintext(cls, flags, valid_until, data, **kw):
        kw = flags.d
        flags       = cls._serialize_flags(flags, **kw)
        valid_until = cls._serialize_valid_until(valid_until, **kw)
        data        = cls._serialize_data(data, **kw)
        data_len    = cls._serialize_data_len(len(data), **kw)
        return cls._join_strings(dict(
                    flags       = flags,
                    valid_until = valid_until,
                    data_len    = data_len,
                    data        = data
        ))

    @classmethod
    def _create_messages(cls, salt, public_flags, flags,
                         valid_until, data, entropy, **kw):
        visible_message = ''
        ticket_len = None
        if public_flags.encrypt:
            ciphertext = kw['ciphertext']
            salt = cls._serialize_salt(salt, **kw)
            public_flags = cls._serialize_public_flags(public_flags, **kw)

            # length of ticket_len not added ??? 
            ticket_len = len(cls.SECURE_TICKET_PREFIX) +\
                         cls.HASH_LEN                  +\
                         len(salt +  public_flags + ciphertext)
            ticket_len += len(cls._serialize_ticket_len(0))
            ticket_len = cls._serialize_ticket_len(ticket_len, **kw)

            visible_message = cls._join_strings(dict(
                salt         = salt,
                ticket_len   = ticket_len,
                public_flags = public_flags,
                data         = ciphertext
            ))
        else:
            salt         = cls._serialize_salt(salt, **kw)
            public_flags = cls._serialize_public_flags(public_flags, **kw)
            flags        = cls._serialize_flags(flags, **kw)
            valid_until  = cls._serialize_valid_until(valid_until, **kw)
            data         = cls._serialize_data(data, **kw)
            data_len     = cls._serialize_data_len(len(data), **kw)

            ticket_len = len(cls.SECURE_TICKET_PREFIX) +\
                         cls.HASH_LEN                  +\
                         len(salt + public_flags + flags +\
                             valid_until + data_len + data)
            ticket_len += len(cls._serialize_ticket_len(0, **kw))
            ticket_len = cls._serialize_ticket_len(ticket_len, **kw)

            visible_message = cls._join_strings(dict(
                salt         = salt,
                ticket_len   = ticket_len,
                public_flags = public_flags,
                flags        = flags,
                valid_until  = valid_until,
                data_len     = data_len,
                data         = data
            ))
        entropy = cls._serialize_entropy(entropy)
        message = visible_message + salt + entropy
        return (message, visible_message)

    @classmethod
    def _create_message(cls, salt, public_flags, flags, valid_until, data, entropy, **kw):
        t = cls._create_messages(
            salt, public_flags, flags, valid_until, data, entropy, **kw
        )
        return t[0]

    @classmethod
    def _create_visible_message(cls, salt, public_flags, flags, valid_until, data, entropy, **kw):
        t = cls._create_messages(
            salt, public_flags, flags, valid_until, data, entropy, **kw
        )
        return t[1]

    def _load(self, hash_, visible_message):
        self._bytes = self.SECURE_TICKET_PREFIX + hash_ + visible_message
        return self

    def _load_from_parts(self, hash_, salt, public_flags,
             flags, valid_until, data, **kw):
        '''
            Loading a *SecureTicket* assumes a hash to exist.
        '''
        visible_message = self._create_visible_message(
            salt, public_flags, flags, valid_until, data, '', **kw)
        self._load(hash_, visible_message)
        return self

    def _set_bytes(self, bytes):
        self._bytes = bytes

    @classmethod
    def _split_deserialize_ticket_plaintext_strings(cls, joined_strings, names):
        '''
            special version of ByteMapper's _split_deserialize_strings() which
            supplies the flags dictionary, also contained within the
            serialized strings, in deserialization of the other values.
            Also, data_len is put in kwargs passed to deserialize() functions,
            since 'data' is null-padded if encrypted and needs a chomp.
        '''
        def string_start_offset(name):
            return getattr(cls, name.upper()+'_OFFSET')
        #
        # first loop joined_strings to find flags and data_len
        #
        flags = None
        data_len = None
        _joined_strings = joined_strings
        for name in sorted(names, key=string_start_offset):
            len_ = getattr(cls, name.upper()+'_LEN')
            string = _joined_strings[:len_]
            _joined_strings = _joined_strings[len_:]
            if name == 'flags':
                deserialize = getattr(cls, '_deserialize_'+name)
                flags = deserialize(string)
            if name == 'data_len':
                deserialize = getattr(cls, '_deserialize_'+name)
                data_len = deserialize(string)
            if not (flags is None) and not (data_len is None):
                break
        assert not flags is None
        assert not data_len is None
        kw = flags.d
        kw['data_len'] = data_len
        #
        # loop again and deserialize everything using flags and data_len
        #
        sorted_values = []
        _joined_strings = joined_strings
        for name in sorted(names, key=string_start_offset):
            len_ = getattr(cls, name.upper()+'_LEN')
            string = _joined_strings[:len_]
            #
            # encrypted data is null-padded to 16-byte alignment
            #
            #if name == 'data':
            #    string = string[:data_len] 
            #if name == 'data':
            #    print 'DATA:%d'%data_len, repr(string)
            _joined_strings = _joined_strings[len_:]
            deserialize = getattr(cls, '_deserialize_'+name)
            val = deserialize(string, **kw)
            sorted_values.append(val)
        return tuple(sorted_values)

    @classmethod
    def _split_deserialize_plaintext_into_tuple(cls, plaintext):
        parts = ('flags', 'valid_until', 'data_len', 'data')
        (flags, valid_until, data_len, data) = \
            cls._split_deserialize_ticket_plaintext_strings(plaintext, parts)
        #data = data[:data_len]
        return (flags, valid_until, data_len, data)

    @classmethod
    def _split_deserialize_plaintext_into_dict(cls, plaintext):
        (flags, valid_until, data_len, data) = \
            cls._split_deserialize_plaintext_into_tuple(plaintext)
        return dict(flags=flags, valid_until=valid_until,
                    data_len=data_len, data=data)
        
    def _recreate_message(self, entropy):
        # _serialize_entropy() doesn't see flags if ticket is encrypted and
        # thus can't [de]serialize based on flags.
        return self.visible_message + self.salt + self._serialize_entropy(entropy)

    @classmethod
    def _parse_visible_message(cls, b):
        return b[cls.VISIBLE_MESSAGE_OFFSET:]

    @classmethod
    def _parse_ciphertext(cls, b):
        return b[cls.CIPHERTEXT_OFFSET:]

    @classmethod
    def _parse_data(cls, b):
        return b[cls.DATA_OFFSET:]


    ######## serialization ########

    #
    # ticket_len
    #
    @classmethod
    def _serialize_ticket_len(cls, val, **kw):
        return struct.pack('I', val)

    @classmethod
    def _deserialize_ticket_len(cls, val, **kw):
        return struct.unpack('I', val)[0]

    #
    # public_flags
    #
    @classmethod
    def _serialize_public_flags(cls, val, **kw):
        return val._serialize()

    @classmethod
    def _deserialize_public_flags(cls, val, **kw):
        return SecureTicketPublicFlags._deserialize(val)

    #
    # flags
    #
    @classmethod
    def _serialize_flags(cls, val, **kw):
        return val._serialize()

    @classmethod
    def _deserialize_flags(cls, val, **kw):
        return SecureTicketFlags._deserialize(val)

    #
    # valid_until
    #
    @classmethod
    def _serialize_valid_until(cls, val, **kw):
        return struct.pack('I', int(val))

    @classmethod
    def _deserialize_valid_until(cls, val, **kw):
        return struct.unpack('I', val)[0] 

    #
    # data_len
    #
    @classmethod
    def _serialize_data_len(cls, val, **kw):
        return struct.pack('I', int(val))

    @classmethod
    def _deserialize_data_len(cls, val, **kw):
        return struct.unpack('I', val)[0] 

    #
    # data
    #
    @classmethod
    def _serialize_data(cls, val, **kw):
        if kw.get('serialize'):
            val = cls._pickler.dumps(val)
        if kw.get('compress'):
            val = zlib.compress(val)
        return val

    @classmethod
    def _deserialize_data(cls, val, **kw):
        if 'data_len' in kw:
            data_len = kw['data_len']
            if len(val) != data_len:
                val = val[:data_len]
        if kw.get('compress'):
            val = zlib.decompress(val)
        if kw.get('serialize'):
            val = cls._pickler.loads(val)
        return val

    #
    # entropy
    #
    @classmethod
    def _serialize_entropy(cls, val):
        return val

    @classmethod
    def _deserialize_entropy(cls, val):
        return val

    #
    # getters for potentially encrypted attributes
    #
    def _get_public_flags(self):
        #
        # Public_flags should not be magically serialized based on flags
        # contents here, as flags is not available here in encrypted tickets.
        #
        return self.__class__._deserialize_public_flags(self.raw_public_flags)

    def _get_flags(self):
        if self.public_flags.encrypt:
            msg = 'flags is encrypted, use SecureTicketService(key)(ticket).flags'
            msg += ' or SecureTicketService(key).get_flags(ticket) instead.'
            raise SecureTicketException(msg)
        return self.__class__._deserialize_flags(self.raw_flags,
                                                 **self.flags.d)

    def _get_valid_until(self):
        if self.public_flags.encrypt:
            msg = 'valid_until is encrypted, use SecureTicketService(key)(ticket).valid_until'
            msg += ' or SecureTicketService(key).get_valid_until(ticket) instead.'
            raise SecureTicketException(msg)
        return self.__class__._deserialize_valid_until(self.raw_valid_until,
                                                       **self.flags.d)

    def _get_data_len(self):
        if self.public_flags.encrypt:
            msg = 'data_len is encrypted, use SecureTicketService(key)(ticket).data_len'
            msg += ' or SecureTicketService(key).get_data_len(ticket) instead.'
            raise SecureTicketException(msg)
        return self.__class__._deserialize_data_len(self.raw_data_len,
                                                    **self.flags.d)

    def _get_data(self):
        if self.public_flags.encrypt:
            msg = 'data is encrypted, use SecureTicketService(key)(ticket).data'
            msg += ' or SecureTicketService(key).get_data(ticket) instead.'
            raise SecureTicketException(msg)
        return self.__class__._deserialize_data(self.raw_data, 
                                                **self.flags.d)


    #
    # properties
    #
    bytes             = property(lambda self : self._bytes)
    visible_message   = property(lambda self : self.__class__._parse_visible_message(self._bytes))
    raw_public_flags  = property(lambda self : self.__class__._parse_public_flags(self._bytes))
    raw_flags         = property(lambda self : self.__class__._parse_flags(self._bytes))
    raw_valid_until   = property(lambda self : self.__class__._parse_valid_until(self._bytes))
    raw_data_len      = property(lambda self : self.__class__._parse_data_len(self._bytes))
    raw_data          = property(lambda self : self.__class__._parse_data(self._bytes))
    data              = property(_get_data)
    ciphertext        = property(lambda self : self.__class__._parse_ciphertext(self._bytes))

    is_encrypted      = property(lambda self : self.public_flags.encrypt)


    def _looks_like_ticket(self):
        return self.prefix == self.SECURE_TICKET_PREFIX

    #
    # public methods
    #
    def tostring(self):
        return self._bytes

    def tobase64(self):
        return base64.b64encode(self.tostring())

    @classmethod
    def fromstring(cls, b):
        ticket = cls()
        ticket._set_bytes(b)
        return ticket

    @classmethod
    def frombase64(cls, b):
        return cls.fromstring(base64.b64decode(b))






class SecureTicketService(object):
    '''
    Some examples:

    Create a random key (use outcommented code):

    >>> #key = SecureTicketService.create_random_key()
    >>> key = 'make doctests work by using static key'

    Lets serialize some data and store in a ticket:

    >>> sts = SecureTicketService(key, serialize=1)
    >>> ticket = sts.create_ticket(['asd', 123, True, None, ['yay']])
    >>> assert type(ticket.data) is list
    >>> assert ticket.data[0] == 'asd'
    >>> assert ticket.data[4][0] == 'yay'

    Lets store a huge text payload with and without compression. First we
    disable serialization which means that only string-type *data* is allowed.

    >>> sts.options.set_serialize(0)
    >>> data = 'A' * 10**5 # 100k
    >>> large = sts.create_ticket(data)
    >>> len(large.data)
    100000
    >>> len(large.tostring())
    100054
    >>> small = sts.create_ticket(data, compress=1)
    >>> len(small.data)
    100000
    >>> len(small.tostring())
    175
    >>> large.data[:5] == small.data[:5] == 'AAAAA'
    True

    What if we want to give a ticket containing secret information to a third
    party, without revealing the ticket contents? By default, the data field is
    in plaintext (albeit protected from manipulation thanks to hashing with our
    secret key). Concider the simple example below:

    >>> plaintext = 'my precious'
    >>> ticket = sts.create_ticket(plaintext)
    >>> ticket.tostring().find('my precious')
    54

    If we enable encryption, all information in ticket is encrypted, including
    metadata, except for the bit flag marking the ticket as encrypted and the
    ticket length field. Encryption algorithm is AES128-CBC from pyCrypto
    (which must be installed) and the key used is the *key* supplied during
    *SecureTicketService* instantiation.

    Data can still be transparently accessed using a *SecureTicketService*
    instance.

    >>> sts.options.set_encrypt(1)
    >>> ticket = sts.create_ticket(plaintext)
    >>> ticket.tostring().find('my precious')
    -1
    >>> sts(ticket).data
    'my precious'

    To enable all features upon *SecureTicketService* instantiation, simply
    use:

    >>> sts = SecureTicketService(key, serialize=1, encrypt=1, compress=1)
    '''

    DEFAULT_TICKET_LIFESPAN = 1200 # 1200 seconds, 20 minutes

    def __init__(self, key, session='', ticket_class=SecureTicket, **options):
        self.options         = self.SecureTicketServiceOptions(**options)
        self._key            = key
        self._session        = session
        self._ticket_class   = ticket_class
        self._flags          = SecureTicketFlags()
        self._public_flags   = SecureTicketPublicFlags()
        self.ticket_lifespan = self.DEFAULT_TICKET_LIFESPAN
        self._flags.sync(self.options)
        self._public_flags.sync(self.options)
        hashfunc           = getattr(hashlib, ticket_class.HASH_TYPE)
        self._hmac_factory = lambda : hmac.HMAC(key, msg='', digestmod=hashfunc)

    class SecureTicketProxy(object):
        def __init__(self, sts, ticket):
            self._sts = sts
            self._ticket = ticket
        hash          = property(lambda self : self._ticket.hash)
        salt          = property(lambda self : self._ticket.salt)
        ticket_len    = property(lambda self : self._ticket.ticket_len)
        public_flags  = property(lambda self : self._ticket.public_flags)
        flags         = property(lambda self : self._sts.get_flags(self._ticket))
        valid_until   = property(lambda self : self._sts.get_valid_until(self._ticket))
        data_len      = property(lambda self : self._sts.get_data_len(self._ticket))
        data          = property(lambda self : self._sts.get_data(self._ticket))

    class SecureTicketHelper(object):
        pass

    class SecureTicketEncryptor(SecureTicketHelper):
        def __init__(self, encrypt_func):
            self._encrypt_func = encrypt_func
        def __call__(self, plaintext):
            return self._encrypt_func(plaintext)

    class SecureTicketHasher(SecureTicketHelper):
        def __init__(self, hash_func):
            self._hash_func = hash_func
        def __call__(self, message):
            return self._hash_func(message)

    SecureTicketServiceOptions = tickets.flags.FlagsFactory(
        'SecureTicketServiceOptions',
        32,
        [
            ('encrypt',           False),
            ('serialize',         False),
            ('compress',          False),
        ]
    )
    #SecureTicketServiceOptions._ignore_invalid_names_on_init = True

    @classmethod
    def _new_entropy(self, size):
        '''
            Do not use _new_entropy() to create entropy for *create_ticket()*.
            The _new_entropy() method returns random junk while *entropy* in
            *create_ticket()* is intended for information.
        '''
        entropy = ''
        while len(entropy) < size:
            entropy += uuid.uuid4().bytes
        return entropy[:size]

    @classmethod
    def create_random_key(cls):
        '''
            Create a key suitable for use in *SecureTicketSession*.
        '''
        return cls._new_entropy(32)

    @classmethod
    def create_random_session(cls):
        '''
            Create a session variable suitable for use in *SecureTicketSession*.
        '''
        return cls._new_entropy(32)


    def __call__(self, ticket):
        return self.SecureTicketProxy(self, ticket)

    def _new_salt(self):
        return self._new_entropy(self._ticket_class.SALT_LEN)

    def _new_hasher(self):
        hmac = self._hmac_factory()
        hash_len = self._ticket_class.HASH_LEN
        def hasher(message):
            hmac.update(message)
            return hmac.digest()[:hash_len]
        return self.SecureTicketHasher(hasher)

    def _hash_message(self, message):
        hasher = self._new_hasher()
        hash_ = hasher(message)
        assert len(hash_) == self._ticket_class.HASH_LEN
        return hash_

    def _encrypt(self, salt, plaintext):
        saes = tickets.crypto.AES(self._key)
        return saes.encrypt(salt, plaintext)

    def _decrypt_first_block(self, ticket):
        '''
            Like _decrypt_ticket_into_tuple() but faster as only one block is
            decrypted. The drawback is that at most 7 bytes of data is returned.

            returns (flags, valid_until, data_len, possibly_partial_data) 
        '''
        aes = tickets.crypto.AES(self._key)
        block = aes.decrypt_first_block(self.salt, ticket.ciphertext)
        (flags, valid_until, data_len, data) =\
            ticket.__class__._split_deserialize_plaintext_into_tuple(block)
        return (flags, valid_until, data_len, data)

    def _decrypt_plaintext_from_ticket(self, ticket):
        aes = tickets.crypto.AES(self._key)
        plaintext = aes.decrypt(ticket.salt, ticket.ciphertext)
        return plaintext

    def _decrypt_ticket_into_tuple(self, ticket):
        '''
            returns decrypted deserialized tuple with the following values:
            (flags, valid_until, data_len, data)
        '''
        plaintext = self._decrypt_plaintext_from_ticket(ticket)
        return ticket.__class__._split_deserialize_plaintext_into_tuple(plaintext)

    def _decrypt_ticket_into_dict(self, ticket):
        '''
            returns decrypted deserialized dict with the following keys:
            'flags', 'valid_until', 'data_len', 'data'
        '''
        plaintext = self._decrypt_plaintext_from_ticket(ticket)
        return ticket.__class__._split_deserialize_plaintext_into_dict(plaintext)


    #
    # decrypt specific parts
    #
    def _decrypt_flags(self, ticket):
        (flags, valid_until, data_len) = \
            self._decrypt_first_block(ticket)
        return flags

    def _decrypt_valid_until(self, ticket):
        (flags, valid_until, data_len) = \
            self._decrypt_first_block(ticket)
        return valid_until

    def _decrypt_data_len(self, ticket):
        (flags, valid_until, data_len) = \
            self._decrypt_first_block(ticket)
        return data_len

    def _decrypt_data(self, ticket):
        (flags, valid_until, data_len, data) = \
            self._decrypt_ticket_into_tuple(ticket)
        return data


    def _add_session_to_entropy(self, salt, entropy):
        #
        # avoid collisions by putting salt between session and entropy
        #
        if self._session:
            entropy = self._session + salt + entropy
        return entropy

    def _new_encryptor(self, iv):
        aes = tickets.crypto.AES(self._key)
        encrypt = lambda plaintext : aes.encrypt(iv, plaintext)
        return self.SecureTicketEncryptor(encrypt)

    def _new_valid_until(self):
        return int(time.time() + self.ticket_lifespan)

    def _create_ticket(self, ticket_class, salt, public_flags, flags,
                       valid_until, data, entropy):
        kw = flags.d
        entropy = self._add_session_to_entropy(salt, entropy)
        if public_flags.encrypt:
            plaintext = ticket_class._create_plaintext(
                flags, valid_until, data, **kw
            )
            flags = valid_until = data = None
            encrypt = self._new_encryptor(salt)
            ciphertext = encrypt(plaintext)
            kw['ciphertext'] = ciphertext
        #
        # Note that the contents of 'flags' are leaked to _create_messages
        # through kw, even if 'flags' is encrypted within ciphertext.
        #
        message, visible_message = ticket_class._create_messages(
            salt, public_flags, flags, valid_until, data, entropy, **kw
        )
        hash_ = self._hash_message(message)

        ticket = ticket_class()._load(hash_, visible_message)
        return ticket

    #
    # user interface
    #

    def set_default_public_flags(self, flags):
        self._public_flags = flags

    def set_default_flags(self, flags):
        self._flags = flags

    def create_ticket(self, data='', entropy='', **options):
        '''
            Create a *SecureTicket* containing *data*. If *entropy* is specified,
            the same *entropy* must be supplied when using *validate_ticket()*,
            or else validation will fail.

            Options can be used to override *SecureTicketService* default
            options.
        '''
        options = self.options.copy_update(**options)

        ticket_class = self._ticket_class
        salt         = self._new_salt()
        public_flags = self._public_flags.copy_sync(options)
        flags        = self._flags.copy_sync(options)
        valid_until  = self._new_valid_until()
        ticket = self._create_ticket(ticket_class, salt, public_flags, flags,
                                      valid_until, data, entropy)
        return ticket

    def validate_ticket(self, ticket, entropy=''):
        '''
            Validate the integrity of *ticket* by recreating its hash and
            verifying that it matches ``ticket.hash``. The validity of
            *ticket*'s lifespan is also verified.
            The 'entropy' argument must match what was used in the call to
            create_ticket().
        '''
        ticket_class = self._ticket_class
        #
        if not ticket._looks_like_ticket():
            msg = 'Ticket broken, invalid prefix: {0}'
            msg = msg.format(repr(ticket.prefix))
            raise InvalidTicketException(msg)
        if not ticket.ticket_len == len(ticket._bytes):
            msg = 'Ticket broken, ticket.ticket_len({0}) != {1}'
            msg = msg.format(ticket.ticket_len, len(ticket._bytes))
            raise InvalidTicketException(msg)
        now = time.time()
        if entropy is None:
            entropy = ''
        entropy = self._add_session_to_entropy(ticket.salt, entropy)
        message = ticket._recreate_message(entropy)
        if self._hash_message(message) != ticket.hash:
            return False
        if self(ticket).valid_until < now:
            return False
        return True


    #
    # ticket attributes; transparent decryption
    #
    def get_flags(self, ticket):
        '''
            Get ``ticket.flags`` and transparently decrypt if necessary.
            Synonymous to ``sts(ticket).flags``, with *sts* being the *SecureTicketService*.
        '''
        flags = None
        if ticket.public_flags.encrypt:
            flags = self._decrypt_ticket_into_tuple(ticket)[0]
        else:
            flags = ticket.flags
        return flags

    def get_valid_until(self, ticket):
        '''
            Get ``ticket.valid_until`` and transparently decrypt if necessary.
            Synonymous to ``sts(ticket).valid_until``, with *sts* being the *SecureTicketService*.
        '''
        valid_until = None
        if ticket.public_flags.encrypt:
            valid_until = self._decrypt_ticket_into_tuple(ticket)[1]
        else:
            valid_until = ticket.valid_until
        return valid_until

    def get_data_len(self, ticket):
        '''
            Get ``ticket.data_len`` and transparently decrypt if necessary.
            Synonymous to ``sts(ticket).data_len``, with *sts* being the *SecureTicketService*.
        '''
        data_len = None
        if ticket.public_flags.encrypt:
            data_len = self._decrypt_ticket_into_tuple(ticket)[2]
        else:
            data_len = ticket.data_len
        return data_len

    def get_data(self, ticket):
        '''
            Get ``ticket.data`` and transparently decrypt and gunzip if necessary.
            Synonymous to ``sts(ticket).data``, with *sts* being the *SecureTicketService*.
        '''
        data = None
        if ticket.public_flags.encrypt:
            data = self._decrypt_ticket_into_tuple(ticket)[3]
        else:
            data = ticket.data
        return data






def _test():
    import doctest
    doctest.testmod()


if __name__ == "__main__":
    _test()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright(c) 2011 Krister Hedfors
#
# This file is part of pyTickets.
#
#    pyTickets is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    pyTickets is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with pyTickets.  If not, see <http://www.gnu.org/licenses/>.
#


import array
import unittest

__all__ = [
    'flip_bit',
]



def flip_bit(s, bitpos):
    arr = array.array('c', s)
    bytepos = bitpos / 8
    bitnum  = bitpos % 8
    ch = ord(arr[bytepos])
    ch ^= 1 << bitnum
    arr[bytepos] = chr(ch)
    return arr.tostring()



class Test_utils(unittest.TestCase):
    def test_test(self):
        assert 1==1

    def setUp(self):
        pass

    def test_flip_bit(self):
        s = '\xf0\x0f'
        assert flip_bit(s, 0) == '\xf1\x0f'
        assert flip_bit(s, 1) == '\xf2\x0f'
        assert flip_bit(s, 2) == '\xf4\x0f'
        assert flip_bit(s, 7) == '\x70\x0f'
        assert flip_bit(s, 8+0) == '\xf0\x0e'
        assert flip_bit(s, 8+1) == '\xf0\x0d'
        assert flip_bit(s, 8+2) == '\xf0\x0b'
        assert flip_bit(s, 8+7) == '\xf0\x8f'


def _test():
    import unittest
    unittest.main()

if __name__ == "__main__":
    _test()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright(c) 2011 Krister Hedfors
#
# This file is part of pyTickets.
#
#    pyTickets is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    pyTickets is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with pyTickets.  If not, see <http://www.gnu.org/licenses/>.
#

'''
    *SecureTicketService* is used to create and validate *SecureTickets*.
    *SecureTickets* are light-weight symmetrically signed data sets with
    a limited lifestpan.

    The *key* passed to *SecureTicketService* is the password and the
    security relies heavily on its strength.  It really should be a 32 byte
    random string as you gain integrity AND performance by using a key of
    32 bytes length (it's padded or replaced by SHA256-hashes of itself
    to make it 32 bytes anyways).
    For your convenience, classmethod *create_random_key()* is provided:

    >>> key = SecureTicketService.create_random_key()
    >>> assert len(key) == 32
    >>> sts = SecureTicketService(key)

    A SecureTicket *ticket* which is successfully validated using
    ``SecureTicketService.validate_ticket()`` can only be created by
    someone who has knowledge of *key*. The entire contents of *ticket* is
    securely hashed using *key* and any change to *ticket* breaks the hash
    validation.

    >>> key = 'Io5IpK9ZTsKpG1ybaLCHkOH4kvHaTEO2imHvkqLVn7I='
    >>> sts = SecureTicketService(key.decode('base64'))
    >>> ticket = sts.create_ticket('someData')
    >>> ticket.data
    'someData'
    >>> sts.validate_ticket(ticket)
    True
    >>> sts2 = SecureTicketService('someOtherKey')
    >>> sts2.validate_ticket(ticket)
    False

    **entropy**

    The optional second argument *entropy* to *create_ticket()*, which must be
    a string if supplied, cannot be obtained from a ticket; it's just
    concatinated together with the rest of ticket when the hash is created.
    The same *entropy* value must therefore be used in
    ``SecureTicketService.validate_hash()`` or else validation fails.

    >>> ticket = sts.create_ticket('someKey', 'someEntropy')
    >>> sts.validate_ticket(ticket)
    False
    >>> sts.validate_ticket(ticket, 'someEntropy')
    True

    **session**

    Many use cases for secure tickets involves (or should involve) the concept
    of a session to prevent various types of attacks. The optional second
    argument *session* to *SecureTicketService()* is used in the same manner as
    *entropy*, but is supplied during *SecureTicketService* instantiation
    instead of during ticket creation.

    >>> sts = SecureTicketService(key, 'someSessionIdentifier')
    
    **options**

    Encryption, serialization and compression of *ticket*'s contents is
    optional. Encrypted tickets will have all its data and metadata encrypted
    with the *key* supplied to *SecureTicketService*. Serialization allows
    complex data types in *data* instead of just strings.  Compression
    (zlib) is useful if the *data* argument is inconveniently large.
    Options and their default values:

        * serialize=False
        * encrypt=False
        * compress=False

    Encrypted ticket attributes must be viewed through a *SecureTicketService*
    instance which provide transparent decryption:

    >>> key = SecureTicketService.create_random_key()
    >>> sts = SecureTicketService(key, serialize=1, compress=1, encrypt=1)
    >>> ticket = sts.create_ticket(['asd', 123], 'ee')
    >>> assert sts.get_data(ticket) == sts(ticket).data == ['asd', 123]
'''

__version__ = '0.9.1'

__all__ = []

from tickets.core import SecureTicketException
from tickets.core import SecureTicketValidationException
from tickets.core import InvalidTicketException
from tickets.core import SecureTicketPublicFlags
from tickets.core import SecureTicketFlags
from tickets.core import SecureTicket
from tickets.core import SecureTicketService

__all__ += [
    'SecureTicketException',
    'SecureTicketValidationException',
    'InvalidTicketException',
    'SecureTicketPublicFlags',
    'SecureTicketFlags',
    'SecureTicket',
    'SecureTicketService',
]

from tickets.web.forms import FormTicketService

__all__ += [
    'FormTicketService',
]



def _test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    _test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright(c) 2011 Krister Hedfors
#
# This file is part of pyTickets.
#
#    pyTickets is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    pyTickets is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with pyTickets.  If not, see <http://www.gnu.org/licenses/>.
#

'''
FormTicketService, FormTicket
=============================
*FormTicketService* provides simple form validation based on
*SecureTicketService*.

Create a form ticket:

>>> key = FormTicketService.create_random_key()
>>> session = 'CurrentUserWebSession'
>>> fts = FormTicketService(key, session)
>>> action = '/some/action'
>>> param_names = ['foo', 'bar']
>>> ticket = fts.create_ticket(action, param_names)
>>> input_params = {'foo':'123'}

Ticket validation is controlled with options. Options (and their
default values) include:
    * *require_form_action=True*
    * *require_param_names=True*
    * *require_param_values=True*
    * *allow_undeclared_param_names=False*

If True, *require_form_action* requires the *form_action* specified upon ticket creation
to match the *form_action* passed to *validate_form*.

If True, *require_param_names* requires all parameter names in both *param_names* and
*params* passed to *create_ticket()* to be available in the *params* argument
passed to *validate_form()*.

If True, *require_param_values* requires param name and value pairs passed in *params*
to *create_ticket()* to exist in *params* passed to *validate_ticket()*.

If True, *allow_undeclared_param_names* allows other param names than what was
specified in *param_names* and *params* in call to *create_ticket()* to be passed to
*validate_form()*.

If any of these validations fail, one of the following exceptions will be
raised, containing a dict which exactly specifies what was passed and which params were
invalid:
    * *InvalidFormActionError*
    * *MissingParamNamesError*
    * *InvalidParamValuesError*
    * *UndeclaredParamNamesError*

'''


__all__ = [
    'FormTicketFlags',
    'FormTicket',
    'FormTicketService',

    'FormTicketValidationError',
    'InvalidFormActionError',
    'MissingParamNamesError',
    'InvalidParamValuesError',
    'UndeclaredParamNamesError',
]




import unittest
import doctest

import tickets
from tickets import SecureTicketValidationException
from tickets import SecureTicket
from tickets import SecureTicketService

import operator
import base64
#import pickle
import cPickle



FormTicketFlags =\
        tickets.flags.FormTicketFlags =\
        tickets.flags.FlagsFactory(
    'FormTicketFlags',
    8,
    [
        ("require_form_action",          True),
        ("require_param_names",          True),
        ("require_param_values",         True),
        ("allow_undeclared_param_names", False),
    ]
)


class FormTicketValidationException(SecureTicketValidationException): pass
class FormTicketValidationError(FormTicketValidationException): pass

class InvalidFormActionError(FormTicketValidationError): pass
class MissingParamNamesError(FormTicketValidationError): pass
class InvalidParamValuesError(FormTicketValidationError): pass
class UndeclaredParamNamesError(FormTicketValidationError): pass



class FormTicket(SecureTicket):
    def __str__(self):
        s = "FormTicket('" + self.tostring() + "')"
        return s
    ftflags              = property(lambda self : self.data[0])
    required_form_action = property(lambda self : self.data[1])
    required_param_names = property(lambda self : self.data[2])
    required_params      = property(lambda self : self.data[3])



class _FormTicketValidators(SecureTicketService):

    def _validate_required_form_action(self, ticket, action_to_validate):
        if ticket.required_form_action != action_to_validate:
            msg = 'Invalid form action: {0}'
            msg = msg.format(action_to_validate)
            err = dict( message=msg,
                        received=action_to_validate,
                        expected=ticket.required_form_action)
            raise InvalidFormActionError(err)
        return True

    def _validate_required_param_names(self, ticket, params_to_validate):
        missing_names = []
        for name in ticket.required_param_names:
            if not name in params_to_validate:
                missing_names.append(name)
        if missing_names:
            msg = 'Missing param names: {0}'
            msg = msg.format(', '.join(missing_names))
            err = dict( message=msg,
                        received=params_to_validate.keys(),
                        expected=ticket.required_param_names,
                        missing=tuple(missing_names))
            raise MissingParamNamesError(err)
        return True

    def _validate_required_param_values(self, ticket, params_to_validate):
        missing_names = []
        invalid_params = {}
        for (name, value) in ticket.required_params:
            if not name in params_to_validate:
                missing_names.append(name)
            if params_to_validate[name] != value:
                invalid_params[name] = params_to_validate[name]
        if missing_names:
            msg = 'Missing param names: {0}'
            msg = msg.format(', '.join(missing_names))
            err = dict( message=msg,
                        received=params_to_validate.keys(),
                        expected=dict(ticket.required_params).keys(),
                        missing=tuple(missing_names))
            raise MissingParamNamesError(err)
        if invalid_params:
            msg = 'Invalid param values: {0}'
            msg = msg.format(repr(invalid_params))
            err = dict( message=msg,
                        received=invalid_params,
                        expected=dict(ticket.required_params),
                        invalid=invalid_params)
            raise InvalidParamValuesError(err)
        return True

    def _validate_declared_param_names(self, ticket, params_to_validate):
        if len(params_to_validate) > len(ticket.required_param_names):
            names_to_validate    = params_to_validate.keys()
            required_param_names = ticket.required_param_names
            undeclared_names = set(names_to_validate) - set(required_param_names)
            undeclared_names = tuple(undeclared_names)
            msg = 'Undeclared param names received: {0}'
            msg = msg.format(tuple(undeclared_names))
            err = dict( message=msg,
                        received=undeclared_names,
                        expected=dict(ticket.required_params),
                        undeclared=undeclared_names)
            raise UndeclaredParamNamesError(err)
        return True



class FormTicketService(_FormTicketValidators):

    def __init__(self, key, session):
        options = dict(
            encrypt   = False,
            serialize = True,
            compress  = False,
        )
        SecureTicketService.__init__(
            self, key, session, ticket_class=FormTicket, **options
        )
        self.ftflags = FormTicketFlags()

    #
    # private
    #
    def _create_data(self, ftflags, form_action, param_names, indexed_param_values):
        return (ftflags, form_action, param_names, indexed_param_values)

    def create_ticket(self,
                      form_action='',
                      param_names=[],
                      params=[],
                      **options):
        '''
            fixme
        '''
        ftflags = self.ftflags.copy_update(**options)
        #
        if type(params) is dict:
            params = sorted(params.items())
        param_names = set(param_names)
        param_names |= set([t[0] for t in params])
        param_names = tuple(sorted(param_names))
        data = self._create_data(ftflags, form_action, param_names, params)
        entropy = ''
        return SecureTicketService.create_ticket(self, data, entropy)


    def validate_form(self, ticket, action_to_validate, params_to_validate):
        if type(params_to_validate) is list:
            params_to_validate = dict(params_to_validate)

        entropy = ''
        if not SecureTicketService.validate_ticket(self, ticket, entropy):
            return False
        ftflags = ticket.ftflags
        if ftflags.require_form_action:
            self._validate_required_form_action(ticket, action_to_validate)

        if ftflags.require_param_names:
            self._validate_required_param_names(ticket, params_to_validate)

        if not ftflags.allow_undeclared_param_names:
            self._validate_declared_param_names(ticket, params_to_validate)
            
        if ftflags.require_param_values:
            self._validate_required_param_values(ticket, params_to_validate)
        #print 'success'
        return True





class Test_FormTicketService(unittest.TestCase):
    def _test_test(self):
        assert 1==2

    def test_all_validators(self):
        session = 'JSESSASDQWE'
        fts = FormTicketService('some key', session)

        action = '/some/action'
        names  = ('aa', 'bb', 'foo')
        params = dict(aa='AA', bb='BB')

        params_good     = dict(aa='AA', bb='BB', foo='FOO')
        params_badval   = dict(aa='AA', bb='XX', foo='FOO')
        params_missing  = dict(aa='AA', bb='BB')
        params_extra    = dict(aa='AA', bb='BB', foo='FOO', bar='BAR')
        action_bad      = '/evil/action'

        #
        # default: strictest validation
        #
        ticket = fts.create_ticket(action, names, params)
        assert       fts.validate_form(ticket, action, params_good)
        self.assertRaises( InvalidFormActionError,
            lambda : fts.validate_form(ticket, action_bad, params_good))
        self.assertRaises( InvalidParamValuesError,
            lambda : fts.validate_form(ticket, action, params_badval))
        self.assertRaises( MissingParamNamesError,
            lambda : fts.validate_form(ticket, action, params_missing))
        self.assertRaises( UndeclaredParamNamesError,
            lambda : fts.validate_form(ticket, action, params_extra))

        #
        # require_form_action=False
        #
        ticket = fts.create_ticket(action, names, params,
            require_form_action=False)

        assert       fts.validate_form(ticket, action, params_good)
        assert       fts.validate_form(ticket, action_bad, params_good)
        self.assertRaises( InvalidParamValuesError,
            lambda : fts.validate_form(ticket, action, params_badval))
        self.assertRaises( MissingParamNamesError,
            lambda : fts.validate_form(ticket, action, params_missing))
        self.assertRaises( UndeclaredParamNamesError,
            lambda : fts.validate_form(ticket, action, params_extra))

        #
        # require_param_names=False
        #
        ticket = fts.create_ticket(action, names, params,
            require_param_names=False)

        assert       fts.validate_form(ticket, action, params_good)
        self.assertRaises( InvalidFormActionError,
            lambda : fts.validate_form(ticket, action_bad, params_good))
        self.assertRaises( InvalidParamValuesError,
            lambda : fts.validate_form(ticket, action, params_badval))
        assert       fts.validate_form(ticket, action, params_missing)
        self.assertRaises( UndeclaredParamNamesError,
            lambda : fts.validate_form(ticket, action, params_extra))

        #
        # require_param_values=False
        #
        ticket = fts.create_ticket(action, names, params,
            require_param_values=False)

        assert     fts.validate_form(ticket, action, params_good)
        self.assertRaises( InvalidFormActionError,
            lambda : fts.validate_form(ticket, action_bad, params_good))
        assert       fts.validate_form(ticket, action, params_badval)
        self.assertRaises( MissingParamNamesError,
            lambda : fts.validate_form(ticket, action, params_missing))
        self.assertRaises( UndeclaredParamNamesError,
            lambda : fts.validate_form(ticket, action, params_extra))

        #
        # allow_undeclared_param_names=True
        #
        ticket = fts.create_ticket(action, names, params,
            allow_undeclared_param_names=True)

        assert       fts.validate_form(ticket, action, params_good)
        self.assertRaises( InvalidFormActionError,
            lambda : fts.validate_form(ticket, action_bad, params_good))
        self.assertRaises( InvalidParamValuesError,
            lambda : fts.validate_form(ticket, action, params_badval))
        self.assertRaises( MissingParamNamesError,
            lambda : fts.validate_form(ticket, action, params_missing))
        assert       fts.validate_form(ticket, action, params_extra)




def _test():
    doctest.testmod()
    unittest.main()

if __name__ == "__main__":
    _test()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright(c) 2011 Krister Hedfors
#
# This file is part of pyTickets.
#
#    pyTickets is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    pyTickets is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with pyTickets.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = [
    'debug'
    'info'
    'warning'
    'error'
    'critical'
    'log'
]


import logging

FORMAT = '%(filename)s:%(lineno)d  %(message)s'

logging.basicConfig(
    format=FORMAT
)

debug    = logging.debug
info     = logging.info
warning  = logging.warning
error    = logging.error
critical = logging.critical
log      = logging.log


def _test():
    import doctest
    import unittest
    doctest.testmod()
    unittest.main()

if __name__ == "__main__":
    _test()




#!/usr/bin/env python
#
# This file is part of pyTickets.
#
#    pyTickets is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    pyTickets is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with pyTickets.  If not, see <http://www.gnu.org/licenses/>.
#
import os
from setuptools import setup

import tickets

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "pytickets",
    version = tickets.__version__,
    author = "Krister Hedfors",
    author_email = "krister@tripleaes.com",
    description = ("pyTickets are light-weight symmetrically signed data"
                   " containers with optional encryption, serialization and"
                   " compression of their contents."
    ),
    license = "GPL",
    keywords = "ticket tickets encrypt secure sign signed ignature hash hashed",
    url = "http://pytickets.tripleaes.com/",
    packages=[
        'tickets',
        'tickets.web',
        'tickets.tests'
    ],
    long_description=tickets.__doc__,
    classifiers=[
        "Intended Audience :: Developers",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python",
        "Topic :: Security :: Cryptography",
    ],
)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright(c) 2011 Krister Hedfors
#
# This file is part of pyTickets.
#
#    pyTickets is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    pyTickets is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with pyTickets.  If not, see <http://www.gnu.org/licenses/>.
#


__all__ = [
    'Test_SimpleAES',
]


import unittest
import sys
#from os.path import abspath, dirname
#sys.path.insert(0, 
#    dirname(dirname(abspath(__file__)))
#)


from tickets.utils import SimpleAES


class Test_SimpleAES(unittest.TestCase):
    def test_test(self):
        assert 1==1
    
    def randstring(self, n):
        import random
        return ''.join(chr(random.randint(0,255)) for _ in xrange(n))

    def test_encrypt_decrypt_various_lengths(self):
        '''
            Test SimepleAES encrypt/decrypt on random strings of len 0 to 1023
        '''
        import uuid
        saes = SimpleAES(uuid.uuid4().bytes)
        for n in xrange(1024):
            text = self.randstring(n)
            salt1 = uuid.uuid4().bytes
            salt2 = uuid.uuid4().bytes
            ciphertext = saes.encrypt(salt1, text)
            assert ciphertext != saes.encrypt(salt2, text)
            assert text == saes.decrypt(salt1, ciphertext)
        

if __name__ == "__main__":
    unittest.main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright(c) 2011 Krister Hedfors
#
# This file is part of pyTickets.
#
#    pyTickets is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    pyTickets is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with pyTickets.  If not, see <http://www.gnu.org/licenses/>.
#


__all__ = [
    'BaseAES',
    'AES',
]

import hashlib
import unittest
import uuid


class SimpleAESException(Exception): pass


class SimpleAES:
    '''
        SimpleAES(key)
    '''




class BaseAES:
    '''
    BaseAES provides a high-level interface to encryption using AES from pyCrypto.

        encrypt(iv, plaintext)
        decrypt(iv, ciphertext)

        Encrypt/decrypt text using AES128 CBC and an IV structure specific
        to SimpleAES. An 8-byte salt must be supplied in each call to
        encrypt().

    '''
    def __init__(self, key):
        import hashlib
        from Crypto.Cipher import AES
        self._AES = AES
        self._key = self._fix_key(key)

    def _new_aes(self, iv):
        iv = iv[:16]
        assert len(iv) == 16
        return self._AES.new(self._key, self._AES.MODE_CBC, iv)

    def _fix_key(self, key):
        while len(key) < 32:
            key += hashlib.sha256(key).digest()
        key = key[:32]
        return key

    @classmethod
    def new_key(cls):
        '''
            Create a key suitable for AES encryption.
        '''
        return uuid.uuid4().bytes + uuid.uuid4().bytes

    @classmethod
    def new_iv(cls):
        '''
            Create a 16-byte IV (128 bits) suitable as an initialization vector
            for AES128 CBC encryption.
        '''
        return uuid.uuid4().bytes

    def iter_encrypt_blocks(self, iv, plaintext):
        aes = self._new_aes(iv)
        while plaintext:
            block = plaintext[:16] 
            if len(block) < 16:
                block += '\x00' * (16-len(block))
            block = aes.encrypt(block)
            yield block
            plaintext = plaintext[16:]
        
    def iter_decrypt_blocks(self, iv, ciphertext):
        aes = self._new_aes(iv)
        while ciphertext:
            block = ciphertext[:16]
            block = aes.decrypt(block)
            yield block
            ciphertext = ciphertext[16:]

    def encrypt(self, iv, plaintext):
        '''
            Encrypt 'plaintext' using AES128 CBC.

            Plaintext is padded with '\x00'-bytes if necessary to make the
            last block 16 bytes wide. This means that you have to keep track
            of plaintext's actual length.
        '''
        ciphertext = ''
        for block in self.iter_encrypt_blocks(iv, plaintext):
            ciphertext += block
        return ciphertext

    def decrypt(self, iv, ciphertext):
        '''
            Decrypt 'ciphertext' using AES128 CBC.

            The length of the returned plaintext is always a multiple of the
            block-size which is 16 for AES128.
        '''
        plaintext = ''
        for block in self.iter_decrypt_blocks(iv, ciphertext):
            plaintext += block
        return plaintext


class AES(BaseAES):
    def decrypt_first_block(self, iv, ciphertext):
        assert len(ciphertext) >= 16
        return self._new_aes(iv).decrypt(ciphertext[:16])


class Test_BaseAES(unittest.TestCase):
    def test_test(self):
        assert 1==1

    def test_encrypt(self):
        plaintext = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
        plaintext += plaintext.lower()
        plaintext += '0123456789'
        iv = uuid.uuid4().bytes
        iv2 = uuid.uuid4().bytes
        key = uuid.uuid4().bytes 
        aes = BaseAES(key)
        ciphertext = aes.encrypt(iv, plaintext)
        assert len(plaintext)  == 62
        assert len(ciphertext) == 64
        assert ciphertext.find('12345') == -1
        #print repr(aes.decrypt(iv, ciphertext))
        assert aes.decrypt(iv, ciphertext)[:-2] == plaintext
        assert aes.decrypt(iv2, ciphertext)[:-2] != plaintext
        




def _test():
    unittest.main()

if __name__ == '__main__':
    _test()

